package org.apache.bcel.verifier.structurals;

import org.apache.bcel.generic.ReferenceType;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import java.util.Arrays;
import org.apache.bcel.generic.Type;

public class LocalVariables implements Cloneable
{
    private final Type[] locals;
    
    public LocalVariables(final int localVariableCount) {
        Arrays.fill(this.locals = new Type[localVariableCount], Type.UNKNOWN);
    }
    
    public Object clone() {
        final LocalVariables lvs = new LocalVariables(this.locals.length);
        System.arraycopy(this.locals, 0, lvs.locals, 0, this.locals.length);
        return lvs;
    }
    
    @Override
    public boolean equals(final Object o) {
        if (!(o instanceof LocalVariables)) {
            return false;
        }
        final LocalVariables lv = (LocalVariables)o;
        if (this.locals.length != lv.locals.length) {
            return false;
        }
        for (int i = 0; i < this.locals.length; ++i) {
            if (!this.locals[i].equals(lv.locals[i])) {
                return false;
            }
        }
        return true;
    }
    
    public Type get(final int slotIndex) {
        return this.locals[slotIndex];
    }
    
    public LocalVariables getClone() {
        return (LocalVariables)this.clone();
    }
    
    @Override
    public int hashCode() {
        return this.locals.length;
    }
    
    public void initializeObject(final UninitializedObjectType uninitializedObjectType) {
        for (int i = 0; i < this.locals.length; ++i) {
            if (this.locals[i] == uninitializedObjectType) {
                this.locals[i] = uninitializedObjectType.getInitialized();
            }
        }
    }
    
    public int maxLocals() {
        return this.locals.length;
    }
    
    public void merge(final LocalVariables localVariable) {
        if (this.locals.length != localVariable.locals.length) {
            throw new AssertionViolatedException("Merging LocalVariables of different size?!? From different methods or what?!?");
        }
        for (int i = 0; i < this.locals.length; ++i) {
            this.merge(localVariable, i);
        }
    }
    
    private void merge(final LocalVariables lv, final int i) {
        try {
            if (!(this.locals[i] instanceof UninitializedObjectType) && lv.locals[i] instanceof UninitializedObjectType) {
                throw new StructuralCodeConstraintException("Backwards branch with an uninitialized object in the local variables detected.");
            }
            if (!this.locals[i].equals(lv.locals[i]) && this.locals[i] instanceof UninitializedObjectType && lv.locals[i] instanceof UninitializedObjectType) {
                throw new StructuralCodeConstraintException("Backwards branch with an uninitialized object in the local variables detected.");
            }
            if (this.locals[i] instanceof UninitializedObjectType && !(lv.locals[i] instanceof UninitializedObjectType)) {
                this.locals[i] = ((UninitializedObjectType)this.locals[i]).getInitialized();
            }
            if (this.locals[i] instanceof ReferenceType && lv.locals[i] instanceof ReferenceType) {
                if (!this.locals[i].equals(lv.locals[i])) {
                    final Type sup = ((ReferenceType)this.locals[i]).getFirstCommonSuperclass((ReferenceType)lv.locals[i]);
                    if (sup == null) {
                        throw new AssertionViolatedException("Could not load all the super classes of '" + this.locals[i] + "' and '" + lv.locals[i] + "'.");
                    }
                    this.locals[i] = sup;
                }
            }
            else if (!this.locals[i].equals(lv.locals[i])) {
                this.locals[i] = Type.UNKNOWN;
            }
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    public void set(final int slotIndex, final Type type) {
        if (type == Type.BYTE || type == Type.SHORT || type == Type.BOOLEAN || type == Type.CHAR) {
            throw new AssertionViolatedException("LocalVariables do not know about '" + type + "'. Use Type.INT instead.");
        }
        this.locals[slotIndex] = type;
    }
    
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        for (int i = 0; i < this.locals.length; ++i) {
            sb.append(Integer.toString(i));
            sb.append(": ");
            sb.append(this.locals[i]);
            sb.append("\n");
        }
        return sb.toString();
    }
}
